@@ -18,15 +18,17 @@ import os
 import stat
 import pickle
 import tempfile
-from sys import platform
+import subprocess
+import sys
 
 from .common import ut, TestCase, UNICODE_FILENAMES, closed_tempfile
+from h5py._hl.files import direct_vfd
 from h5py import File
 import h5py
 from .. import h5
-
 import pathlib
 
+
 class TestFileOpen(TestCase):
 
     """
@@ -42,7 +44,6 @@ class TestFileOpen(TestCase):
             with File(fname):
                 pass
 
-
         # Existing readonly file; open read-only
         with File(fname, 'w'):
             pass
@@ -98,6 +99,14 @@ class TestFileOpen(TestCase):
         finally:
             fid.close()
 
+        os.chmod(fname, stat.S_IREAD)  # Make file read-only
+        try:
+            with pytest.raises(PermissionError):
+                File(fname, 'a')
+        finally:
+            # Make it writable again so it can be deleted on Windows
+            os.chmod(fname, stat.S_IREAD | stat.S_IWRITE)
+
     def test_readonly(self):
         """ Mode 'r' opens file in readonly mode """
         fname = self.mktemp()
@@ -135,8 +144,7 @@ class TestFileOpen(TestCase):
         with self.assertRaises(ValueError):
             File(self.mktemp(), 'mongoose')
 
-@ut.skipIf(h5py.version.hdf5_version_tuple < (1, 10, 1),
-               'Requires HDF5 1.10.1 or later')
+
 class TestSpaceStrategy(TestCase):
 
     """
@@ -174,6 +182,84 @@ class TestSpaceStrategy(TestCase):
         dset[...] = 1
         fid.close()
 
+
+@pytest.mark.mpi_skip
+class TestPageBuffering(TestCase):
+    """
+        Feature: Use page buffering
+    """
+
+    def test_only_with_page_strategy(self):
+        """Allow page buffering only with fs_strategy="page".
+        """
+        fname = self.mktemp()
+        with File(fname, mode='w', fs_strategy='page', page_buf_size=16*1024):
+            pass
+        with self.assertRaises(OSError):
+            File(fname, mode='w', page_buf_size=16*1024)
+        with self.assertRaises(OSError):
+            File(fname, mode='w', fs_strategy='fsm', page_buf_size=16*1024)
+        with self.assertRaises(OSError):
+            File(fname, mode='w', fs_strategy='aggregate', page_buf_size=16*1024)
+
+    def test_check_page_buf_size(self):
+        """Verify set page buffer size, and minimum meta and raw eviction criteria."""
+        fname = self.mktemp()
+        pbs = 16 * 1024
+        mm = 19
+        mr = 67
+        with File(fname, mode='w', fs_strategy='page',
+                  page_buf_size=pbs, min_meta_keep=mm, min_raw_keep=mr) as f:
+            fapl = f.id.get_access_plist()
+            self.assertEqual(fapl.get_page_buffer_size(), (pbs, mm, mr))
+
+    @pytest.mark.skipif(h5py.version.hdf5_version_tuple > (1, 14, 3),
+                        reason='Requires HDF5 <= 1.14.3')
+    def test_too_small_pbs(self):
+        """Page buffer size must be greater than file space page size."""
+        fname = self.mktemp()
+        fsp = 16 * 1024
+        with File(fname, mode='w', fs_strategy='page', fs_page_size=fsp):
+            pass
+        with self.assertRaises(OSError):
+            File(fname, mode="r", page_buf_size=fsp-1)
+
+    @pytest.mark.skipif(h5py.version.hdf5_version_tuple < (1, 14, 4),
+                        reason='Requires HDF5 >= 1.14.4')
+    def test_open_nonpage_pbs(self):
+        """Open non-PAGE file with page buffer set."""
+        fname = self.mktemp()
+        fsp = 16 * 1024
+        with File(fname, mode='w'):
+            pass
+        with File(fname, mode='r', page_buf_size=fsp) as f:
+            fapl = f.id.get_access_plist()
+            assert fapl.get_page_buffer_size()[0] == 0
+
+    @pytest.mark.skipif(h5py.version.hdf5_version_tuple < (1, 14, 4),
+                    reason='Requires HDF5 >= 1.14.4')
+    def test_smaller_pbs(self):
+        """Adjust page buffer size automatically when smaller than file page."""
+        fname = self.mktemp()
+        fsp = 16 * 1024
+        with File(fname, mode='w', fs_strategy='page', fs_page_size=fsp):
+            pass
+        with File(fname, mode='r', page_buf_size=fsp-100) as f:
+            fapl = f.id.get_access_plist()
+            assert fapl.get_page_buffer_size()[0] == fsp
+
+    def test_actual_pbs(self):
+        """Verify actual page buffer size."""
+        fname = self.mktemp()
+        fsp = 16 * 1024
+        pbs = 2 * fsp
+        with File(fname, mode='w', fs_strategy='page', fs_page_size=fsp):
+            pass
+        with File(fname, mode='r', page_buf_size=pbs-1) as f:
+            fapl = f.id.get_access_plist()
+            self.assertEqual(fapl.get_page_buffer_size()[0], fsp)
+
+
 class TestModes(TestCase):
 
     """
@@ -232,6 +318,62 @@ class TestDrivers(TestCase):
         self.assertEqual(fid.driver, 'stdio')
         fid.close()
 
+        # Testing creation with append flag
+        fid = File(self.mktemp(), 'a', driver='stdio')
+        self.assertTrue(fid)
+        self.assertEqual(fid.driver, 'stdio')
+        fid.close()
+
+    @ut.skipUnless(direct_vfd,
+                   "DIRECT driver is supported on Linux if hdf5 is "
+                   "built with the appriorate flags.")
+    def test_direct(self):
+        """ DIRECT driver is supported on Linux"""
+        fid = File(self.mktemp(), 'w', driver='direct')
+        self.assertTrue(fid)
+        self.assertEqual(fid.driver, 'direct')
+        default_fapl = fid.id.get_access_plist().get_fapl_direct()
+        fid.close()
+
+        # Testing creation with append flag
+        fid = File(self.mktemp(), 'a', driver='direct')
+        self.assertTrue(fid)
+        self.assertEqual(fid.driver, 'direct')
+        fid.close()
+
+        # 2022/02/26: hnmaarrfk
+        # I'm actually not too sure of the restriction on the
+        # different valid block_sizes and cbuf_sizes on different hardware
+        # platforms.
+        #
+        # I've learned a few things:
+        #    * cbuf_size: Copy buffer size must be a multiple of block size
+        # The alignment (on my platform x86-64bit with an NVMe SSD
+        # could be an integer multiple of 512
+        #
+        # To allow HDF5 to do the heavy lifting for different platform,
+        # We didn't provide any arguments to the first call
+        # and obtained HDF5's default values there.
+
+        # Testing creation with a few different property lists
+        for alignment, block_size, cbuf_size in [
+                default_fapl,
+                (default_fapl[0], default_fapl[1], 3 * default_fapl[1]),
+                (default_fapl[0] * 2, default_fapl[1], 3 * default_fapl[1]),
+                (default_fapl[0], 2 * default_fapl[1], 6 * default_fapl[1]),
+                ]:
+            with File(self.mktemp(), 'w', driver='direct',
+                      alignment=alignment,
+                      block_size=block_size,
+                      cbuf_size=cbuf_size) as fid:
+                actual_fapl = fid.id.get_access_plist().get_fapl_direct()
+                actual_alignment = actual_fapl[0]
+                actual_block_size = actual_fapl[1]
+                actual_cbuf_size = actual_fapl[2]
+                assert actual_alignment == alignment
+                assert actual_block_size == block_size
+                assert actual_cbuf_size == actual_cbuf_size
+
     @ut.skipUnless(os.name == 'posix', "Sec2 driver is supported on posix")
     def test_sec2(self):
         """ Sec2 driver is supported on posix """
@@ -240,6 +382,12 @@ class TestDrivers(TestCase):
         self.assertEqual(fid.driver, 'sec2')
         fid.close()
 
+        # Testing creation with append flag
+        fid = File(self.mktemp(), 'a', driver='sec2')
+        self.assertTrue(fid)
+        self.assertEqual(fid.driver, 'sec2')
+        fid.close()
+
     def test_core(self):
         """ Core driver is supported (no backing store) """
         fname = self.mktemp()
@@ -249,6 +397,12 @@ class TestDrivers(TestCase):
         fid.close()
         self.assertFalse(os.path.exists(fname))
 
+        # Testing creation with append flag
+        fid = File(self.mktemp(), 'a', driver='core')
+        self.assertTrue(fid)
+        self.assertEqual(fid.driver, 'core')
+        fid.close()
+
     def test_backing(self):
         """ Core driver saves to file when backing store used """
         fname = self.mktemp()
@@ -304,64 +458,31 @@ class TestDrivers(TestCase):
         with self.assertRaises(ValueError):
             File(tf, 'w', driver='core')
 
-
     # TODO: family driver tests
 
 
-@ut.skipUnless(h5py.version.hdf5_version_tuple < (1, 10, 2),
-               'Requires HDF5 before 1.10.2')
-class TestLibver(TestCase):
-
-    """
-        Feature: File format compatibility bounds can be specified when
-        opening a file.
-    """
-
-    def test_default(self):
-        """ Opening with no libver arg """
-        f = File(self.mktemp(), 'w')
-        self.assertEqual(f.libver, ('earliest', 'latest'))
-        f.close()
-
-    def test_single(self):
-        """ Opening with single libver arg """
-        f = File(self.mktemp(), 'w', libver='latest')
-        self.assertEqual(f.libver, ('latest', 'latest'))
-        f.close()
-
-    def test_multiple(self):
-        """ Opening with two libver args """
-        f = File(self.mktemp(), 'w', libver=('earliest', 'latest'))
-        self.assertEqual(f.libver, ('earliest', 'latest'))
-        f.close()
-
-    def test_none(self):
-        """ Omitting libver arg results in maximum compatibility """
-        f = File(self.mktemp(), 'w')
-        self.assertEqual(f.libver, ('earliest', 'latest'))
-        f.close()
-
-
-@ut.skipIf(h5py.version.hdf5_version_tuple < (1, 10, 2),
-           'Requires HDF5 1.10.2 or later')
+@pytest.mark.skipif(
+    h5py.version.hdf5_version_tuple[1] % 2 != 0 ,
+    reason='Not HDF5 release version'
+)
 class TestNewLibver(TestCase):
 
     """
         Feature: File format compatibility bounds can be specified when
         opening a file.
-
-        Requirement: HDF5 1.10.2 or later
     """
 
     @classmethod
     def setUpClass(cls):
-        super(TestNewLibver, cls).setUpClass()
+        super().setUpClass()
 
         # Current latest library bound label
         if h5py.version.hdf5_version_tuple < (1, 11, 4):
             cls.latest = 'v110'
-        else:
+        elif h5py.version.hdf5_version_tuple < (1, 13, 0):
             cls.latest = 'v112'
+        else:
+            cls.latest = 'v114'
 
     def test_default(self):
         """ Opening with no libver arg """
@@ -388,7 +509,7 @@ class TestNewLibver(TestCase):
         f.close()
 
     @ut.skipIf(h5py.version.hdf5_version_tuple < (1, 11, 4),
-           'Requires HDF5 1.11.4 or later')
+               'Requires HDF5 1.11.4 or later')
     def test_single_v112(self):
         """ Opening with "v112" libver arg """
         f = File(self.mktemp(), 'w', libver='v112')
@@ -437,7 +558,6 @@ class TestUserblock(TestCase):
         with self.assertRaises(ValueError):
             File(self.mktemp(), 'w', userblock_size='non')
 
-
     def test_write_only(self):
         """ User block only allowed for write """
         name = self.mktemp()
@@ -548,9 +668,9 @@ class TestUnicode(TestCase):
         Modes 'r' and 'r+' do not create files even when given unicode names
         """
         fname = self.mktemp(prefix=chr(0x201a))
-        with self.assertRaises(IOError):
+        with self.assertRaises(OSError):
             File(fname, 'r')
-        with self.assertRaises(IOError):
+        with self.assertRaises(OSError):
             File(fname, 'r+')
 
 
@@ -620,6 +740,7 @@ class TestClose(TestCase):
         f.close()
         f.close()
 
+
 class TestFlush(TestCase):
 
     """
@@ -686,6 +807,27 @@ class TestCloseInvalidatesOpenObjectIDs(TestCase):
             self.assertFalse(bool(f1.id))
             self.assertFalse(bool(g1.id))
 
+    def test_close_one_handle(self):
+        fname = self.mktemp()
+        with File(fname, 'w') as f:
+            f.create_group('foo')
+
+        f1 = File(fname)
+        f2 = File(fname)
+        g1 = f1['foo']
+        g2 = f2['foo']
+        assert g1.id.valid
+        assert g2.id.valid
+        f1.close()
+        assert not g1.id.valid
+        # Closing f1 shouldn't close f2 or objects belonging to it
+        assert f2.id.valid
+        assert g2.id.valid
+
+        f2.close()
+        assert not f2.id.valid
+        assert not g2.id.valid
+
 
 class TestPathlibSupport(TestCase):
 
@@ -721,7 +863,7 @@ class TestPickle(TestCase):
 # unittest doesn't work with pytest fixtures (and possibly other features),
 # hence no subclassing TestCase
 @pytest.mark.mpi
-class TestMPI(object):
+class TestMPI:
     def test_mpio(self, mpi_file_name):
         """ MPIO driver and options """
         from mpi4py import MPI
@@ -730,8 +872,14 @@ class TestMPI(object):
             assert f
             assert f.driver == 'mpio'
 
-    @pytest.mark.skipif(h5py.version.hdf5_version_tuple < (1, 8, 9),
-        reason="mpio atomic file operations were added in HDF5 1.8.9+")
+    def test_mpio_append(self, mpi_file_name):
+        """ Testing creation of file with append """
+        from mpi4py import MPI
+
+        with File(mpi_file_name, 'a', driver='mpio', comm=MPI.COMM_WORLD) as f:
+            assert f
+            assert f.driver == 'mpio'
+
     def test_mpi_atomic(self, mpi_file_name):
         """ Enable atomic mode for MPIO driver """
         from mpi4py import MPI
@@ -751,8 +899,6 @@ class TestMPI(object):
         f.close()
 
 
-@ut.skipIf(h5py.version.hdf5_version_tuple < (1, 10, 1),
-               'Requires HDF5 1.10.1 or later')
 class TestSWMRMode(TestCase):
 
     """
@@ -782,17 +928,90 @@ class TestSWMRMode(TestCase):
         fid.close()
 
 
-# unittest doesn't work with pytest fixtures (and possibly other features),
-# hence no subclassing TestCase
-class TestROS3:
-    @pytest.mark.skipif(h5py.version.hdf5_version_tuple < (1, 10, 6)
-                        or not h5.get_config().ros3,
-                        reason="ros3 file operations were added in HDF5 1.10.6+")
-    def test_ros3(self):
-        """ ROS3 driver and options """
+@pytest.mark.skipif(
+    h5py.version.hdf5_version_tuple < (1, 12, 1) and (
+    h5py.version.hdf5_version_tuple[:2] != (1, 10) or h5py.version.hdf5_version_tuple[2] < 7),
+    reason="Requires HDF5 >= 1.12.1 or 1.10.x >= 1.10.7")
+@pytest.mark.skipif("HDF5_USE_FILE_LOCKING" in os.environ,
+                    reason="HDF5_USE_FILE_LOCKING env. var. is set")
+class TestFileLocking:
+    """Test h5py.File file locking option"""
 
-        with File("https://dandiarchive.s3.amazonaws.com/ros3test.hdf5", 'r',
-                  driver='ros3') as f:
-            assert f
-            assert 'mydataset' in f.keys()
-            assert f["mydataset"].shape == (100,)
+    def test_reopen(self, tmp_path):
+        """Test file locking when opening twice the same file"""
+        fname = tmp_path / "test.h5"
+
+        with h5py.File(fname, mode="w", locking=True) as f:
+            f.flush()
+
+            # Opening same file in same process without locking is expected to fail
+            with pytest.raises(OSError):
+                with h5py.File(fname, mode="r", locking=False) as h5f_read:
+                    pass
+
+            with h5py.File(fname, mode="r", locking=True) as h5f_read:
+                pass
+
+            if h5py.version.hdf5_version_tuple < (1, 14, 4):
+                with h5py.File(fname, mode="r", locking='best-effort') as h5f_read:
+                    pass
+            else:
+                with pytest.raises(OSError):
+                    with h5py.File(fname, mode="r", locking='best-effort') as h5f_read:
+                        pass
+
+
+    def test_unsupported_locking(self, tmp_path):
+        """Test with erroneous file locking value"""
+        fname = tmp_path / "test.h5"
+        with pytest.raises(ValueError):
+            with h5py.File(fname, mode="r", locking='unsupported-value') as h5f_read:
+                pass
+
+    def test_multiprocess(self, tmp_path):
+        """Test file locking option from different concurrent processes"""
+        fname = tmp_path / "test.h5"
+
+        def open_in_subprocess(filename, mode, locking):
+            """Open HDF5 file in a subprocess and return True on success"""
+            h5py_import_dir = str(pathlib.Path(h5py.__file__).parent.parent)
+
+            process = subprocess.run(
+                [
+                    sys.executable,
+                    "-c",
+                    f"""
+import sys
+sys.path.insert(0, {h5py_import_dir!r})
+import h5py
+f = h5py.File({str(filename)!r}, mode={mode!r}, locking={locking})
+                    """,
+                ],
+                capture_output=True)
+            return process.returncode == 0 and not process.stderr
+
+        # Create test file
+        with h5py.File(fname, mode="w", locking=True) as f:
+            f["data"] = 1
+
+        with h5py.File(fname, mode="r", locking=False) as f:
+            # Opening in write mode with locking is expected to work
+            assert open_in_subprocess(fname, mode="w", locking=True)
+
+
+def test_close_gc(writable_file):
+    # https://github.com/h5py/h5py/issues/1852
+    for i in range(100):
+        writable_file[str(i)] = []
+
+    filename = writable_file.filename
+    writable_file.close()
+
+    # Ensure that Python's garbage collection doesn't interfere with closing
+    # a file. Try a few times - the problem is not 100% consistent, but
+    # normally showed up on the 1st or 2nd iteration for me. -TAK, 2021
+    for i in range(10):
+        with h5py.File(filename, 'r') as f:
+            refs = [d.id for d in f.values()]
+            refs.append(refs)   # Make a reference cycle so GC is involved
+            del refs  # GC is likely to fire while closing the file
